/*
    Program Name:  TBPrint
         Compile:  Clipper TBPrint /n /w /a

     Description: Generic print facility for Clipper TBrowse objects.
  Primary Author: C.W. Kreimer
      Concept by: Ed Stoll
            Date: December 19, 1991

       Copyright: C.W. Kreimer and Associates, Inc.

      Disclaimer: Released into Public domain 07-01-92 for benefit of
                  Clipper 5.01 programmers.  C.W. Kreimer & Associates
                  will not be held liable for any problems arising 
		  from the use of this code.

       Arguments: o, lSkipPaging, lForceToFile
    Return Value: Nil


   //////////////////////////////////////////////////////////////////////
   Modifications:

   1. Description:  Provide the ability to print to a file.

          Concept:  C.W. Kreimer
           Author:  C.W. Kreimer
             Date:  February 24, 1992

   2. Description:  Provide totaling and sub totaling of numeric data.

          Concept:  Ed Stoll & Greg Schmitt
           Author:  Ed Stoll
             Date:  June 15, 1992

    /////////////////////////////////////////////////////////////////////

    Usage:  Within your key handling routine of any TBrowse object include
            calls to TBPrint() as defined below.

         TBPrint( o )            // Will Print the TBrowse to the printer
         TBPrint( o, .f., .t. )  // Will Print the TBrowse to a File.

	 Example:

	 nKey = inkey( 0 )
	 do case
	 	case nKey == K_DOWN
			oBrowse:down() // Move the browse cell down
	 	case nKey == K_UP  
			oBrowse:up()  // Move the browse cell up
	 	case nKey == K_SH_F7
			TBPrint( oBrowse )  // Will Print the TBrowse
					    // to the printer
		// etc.
	 endcase

   //////////////////////////////////////////////////////////////////////
   When Building the TBrowse Object:

   Since the TBrowse Object does not provide for a Report title
   include a call to the DataDict function to provide your own
   report title as shown:

      datadict( o, 'PAGEHEADER', 'Page Title for TBPrint' )

   To inform TBPrint() that you want Totals to be accumulated and
   printed, include the following DataDict call:

      datadict( o, "TB_TOTALS", .t. )      // Turn on Totaling

   The above will provide you with a final total line - in order to
   print sub total information, another call to DataDict with the
   row name of "TB_BREAK" must be made - the last parameter must
   be an array of TBrowse Column numbers in their break order!

   So that given:

      datadict( o, "TB_BREAK", { 1, 2 } )

      TBPrint will know the order to check column values for subtotal
      breaks - but more must be done.  Keep reading.....


   For printing the Total line(s), You usually will want some kind
   of literal string to proceed the data.  This is accomplished with
   a call to DataDict() and pass the row name of "TB_STRINGS".  The
   last parameter again must be an array of Strings with the first
   string being used for the final total, then all others match to the
   column break order defined in the TB_BREAK datadict item.

      datadict( o, "TB_STRINGS", { "Grand Totals:  ", ;
                   "Project Total:  ", "Employee Total:  " } )

   To get a better Sub Total line - you may embed two "#" - these will be
   replace with their current values at print time.  For example, if you
   had a TBrowse using all the DataDict() statements above and in Column 1
   was the Project, and Column 2 had the Employee name, you could use the
   following "TB_STRINGS" definition:

   datadict( o, "TB_STRINGS", { "Grand Totals:  ", ;
                                "Project:  ##", "Employee:  ##" } )

   and get the following subtotal lines:

      Employee:  JOHN               9999     9999

      Project:  SAMPLE              9999     9999

      Grand Totals:                 9999     9999



   All of the above DataDict() calls worked against the TBrowse object, the
   following DataDict() calls work with a Column Object!

   For each column number listed in the array of the "TB_BREAK" DataDict()
   statement above, a corresponding DataDict() statement must be issued
   against the Column object.  This will what to evaluate for control
   breaks.

   The following statement will cause TBPrint to just look at the contents
   of the cell.

      datadict( o:GetColumn( 1 ), "COL_BREAK", Nil )

   The next statement gives us some more flexibility, we will cause a break
   whenever the first three characters change in the field.

      datadict( o:GetColumn( 1 ), "COL_BREAK", ;
                {|| substr( Project, 1, 3 ) } )

   TBPrint does not assume to print all your numeric data.  You must
   inform TBPrint what columns are to be totaled and they MUST be NUMERIC.

   This is accomplished as shown:

      datadict( o:GetColumn( 3 ), "COL_TOTAL", .t. )
      datadict( o:GetColumn( 4 ), "COL_TOTAL", .t. )

   For Decimal positions other than the system default - A column's
   decimal format may be alter with the following statement where the
   last parameter is the number of positions to the right of the period
   to display.

      datadict( o:GetColumn( 4 ), "D", 1 )


   ////////////////////////////////////////////////////////////////////

   That about covers it.  We recommend putting in some screen displays
   to inform the user as to what is happening.

   Also - please refer to the sample program to see how to incorporate
   the DataDict() calls more clearly.


                                                                           */

#include 'box.ch'
#include 'inkey.ch'
#include 'setcurs.ch'
#include 'error.ch'

#define ESC    chr( 27 )

#define USER_PRT_TYPE    1   // Static Array pointer for User printer type
#define USER_PRT_PORT    2   // Static Array pointer for User printer port
#define USER_CHAR_SET    3   // Static Array pointer for User character set
#define USER_NAME        4   // Static Array pointer for User name
#define USER_INFO_COUNT  4   // Static Array size

#Define DECIMALS       "D"   // Browse Print Engine manifest constant for Decimal width
#Define COLSEPS        'C'   // Browse Print Engine Manifest constant for column separators.

// UDC - Default <var> to <value>
#xcommand DEFAULT <p> TO <v> [, <p2> TO <v2> ] =>;
	<p> := if( <p> == NIL, <v>, <p>) ;
	[; <p2> := if( <p2> == NIL, <v2>, <p2>) ]

Static lInLandScape        // Landscape mode flag
Static lBox_Chars := .f.   // Print Box around report flag
Static cBoxChars           // Character string of box characters
Static aUserInfo           // Holds User printing information
Static lTotalFlag          // Informs us to perform total routines or not
Static nWidth              // how wide is the paper being used
Static nCurLine            // Current Print Line
Static aSBreaks            // Array of Control break values for subtotals
Static aOffSet             // Array of column printing positions
Static nAFooterHeight      // Number of lines the footers will take up
Static nMargin             // The Left margin print column number
Static nHowWide            // The width of the current TBrowse
Static nFormSize           // The length of the paper
Static lPageSkip           // lSkipPaging saved as a FW-Static

//************************************************************************
Function TBPrint( o, lSkipPaging, lForceToFile )
   Local i, j, nSkipped := 1, nHeaderHeight := 0, nFooterHeight := 0, ;
         xNoCenter, xBox, nKey, nExitState := 0, cPageHead, cPageFoot
   Local nSecurity, nOldMargin := set( _SET_MARGIN ), ;
         xOldDevice := set( _SET_DEVICE ), xOldCursor := setcursor( SC_NONE )
   Local aColTypes  := array( o:colCount ), xDecWidth, temp, ;
         nDefWidth := datadict( o, DECIMALS ), nThisMargin, ;
         nNumColSep := datadict( o, COLSEPS ), cCurColor, e, w, d
   Local xOldEscape := set( _SET_ESCAPE ), cPtrName, ;
         n_colpos, nTotalLevel := 1, aBreakColumns, aInitial, aTotals
   Local cScreen := SaveScreen( 0, 0, maxrow(), maxcol() )

/*
   Initialize variables to their default/minimum values
                                                               */
   default lSkipPaging to .f., ;             // See UDC Definition above
           lForceToFile to .f.

   aOffSet := array( o:ColCount )   // array of printing positions

   nAFooterHeight := 0           // Initialize the height of the footers
   lPageSkip := lSkipPaging      // Initialize the FW-Static

   // Get the Current User's Printing information
   aUserInfo := SetUserInfo()

   set( _SET_ESCAPE, 'ON' )   // enable escapes

   // If not forced to file - check the users configuration - if
   // user's print port is set to 5 ( file ) - force to file.
   If ! lForceToFile
      // Check for User info set to file
      lForceToFile := ( aUserInfo[ USER_PRT_PORT ] == '5' )
   Endif

   // Default lBox_Chars to false if printing to file
   If lForceToFile
      lBox_Chars = .f.
   Endif

   // If BOX is True - set lBoxChars to .t.
   xBox = datadict( o, 'BOX' )
   If ( xBox <> NIL ) .and. xBox
      lBox_Chars := .t.
   Endif

   If lBox_Chars
      nWidth = 2
      // If the user's configuration can handle the full ascii set
      // use graphical box characters - otherwise use the boring
      // +|- characters
      If aUserInfo[ USER_CHAR_SET ] <> '1'
         // Top Left,Top Right,Bottom Left,Bottom Right, Sides, Top + Bottom
         cBoxChars = chr(201)+chr(187)+chr(200)+chr(188)+chr(186)+chr(205)
      Else
         cBoxChars = '++++|-'
      Endif
   Else
      If lForceToFile
         // Establish a filename based upon the user's login name
         // and the next available number - i.e. JOHN0003.rpt
         cPtrName = ptrname()
         If empty( cPtrName )
            Alert('Unable to create print file.')

            // Force return - unable to create a file.
            Return nil
         Endif
      Endif
      cBoxChars = space( 6 )
      nWidth = 0
   Endif

   // Give some feedback to the user!!!!!!
   @ 9, 20 clear to 16, 60
   @ 10, 21, 15, 59 Box B_DOUBLE+" " 
   Center( 12, "Printing TBrowse Data...")
   Center( 14, "Press ESC to abort.")

   // Are totals to be accumulated in this report  ?????
   // If so - the TBrowse's Cargo slot will have a
   //         a slot called TOTALS containing a logical true.
   // Else  - DataDict() will return Nil

   lTotalFlag := ( DataDict( o, "TB_TOTALS" ) <> Nil )

   If lTotalFlag
      // TB_BREAK - if defined - contains an array of columns to break by
      aBreakColumns := DataDict( o, "TB_BREAK" )

      // Ensure DataDict() returned an array before we try to adjust
      // the variable nTotalLevel
      If ValType( aBreakColumns ) == 'A'
         nTotalLevel += len( aBreakColumns ) // Increment by size of array
      Else
         aBreakColumns := {}
      Endif

      // Initialize an array that can be stored into
      // a column objects cargo slot around line 150 of this program
      aInitial := array( nTotalLevel )
      aFill( aInitial, 0 )
   Endif


   lInLandScape = .f.

   default nDefWidth to 0           // See UDC above
   default nNumColSep to 1

   pagecount( 0 )     // set the page counter to zero
   setprc( 0, 0 )     // set the printer row & col to zero


   // Loop through the columns and set up the printing
   // positions for each into the FW-Static array aOffSet

   for i = 1 to o:colCount
      // calculate the data type
      aColTypes[i] = valtype( eval( o:getcolumn( i ):block ) )

      // calculate the total width
      aOffset[i] = nWidth

      nWidth += ( o:colWidth( i ) + nNumColSep )

      // calc header height
      nHeaderHeight = max( nHeaderHeight,;
               if( o:getcolumn( i ):heading == nil, 0, ;
                  how_many_text_elements( o:getcolumn( i ):heading, ';')))

      // calc footer height
      nFooterHeight = max( nFooterHeight,;
            if( o:getcolumn( i ):heading == nil, 0, ;
                how_many_text_elements( o:getcolumn( i ):footing, ';')))

      // If totaling is requested - initialize an array
      //    in cargo of the column object to contain the totals

      If lTotalFlag
         If aColTypes[ i ] == 'N'      // Must be numeric to be totaled
            // Insert the array in the column objects cargo slot
            // NOTE:  The array is passed via the Clone function
            //        because all arrays are handled referentially!
            DataDict( o:GetColumn( i ), "COLUMN_TOTALS", ;
                      aClone( aInitial ) )
         Endif
      Endif
   next

   nWidth -= nNumColSep

   cPageHead = datadict( o, 'PAGEHEADER' )   // Extract the TBrowses
   cPageFoot = datadict( o, 'PAGEFOOTER' )   // Page Header & Footer

   // If NO_CENTER is set to true - set margin to 0
   // otherwise auto center tbrowse
   xNoCenter = datadict( o, 'NO_CENTER' )

   If ! ( xNoCenter == NIL ) .and. xNoCenter
      nMargin = 0
   Else
      // Calculate margin for centering browse on paper
      Do Case
         Case nWidth < 79
            nMargin := ( 80 - nWidth ) * .5
         Case nWidth > 78 .and. nWidth < 133
            nMargin := ( 132 - nWidth ) * .5
         Case nWidth > 133
            nMargin := ( 178 - nWidth ) * .5
      EndCase

      nMargin = int( nMargin )

      // If nMargin is negative - reset to zero
      nMargin := if( nMargin < 0, 0, nMargin )
   Endif

   // fill up static arrays
   rowHeading( o, nHeaderHeight )
   rowFooting( o, nFooterHeight )

   nHeaderHeight += if( cPagehead == nil, 0, ;
                     how_many_text_elements( cPagehead,';' ))
   nAFooterHeight = if( cPagefoot == nil, 0, ;
                     how_many_text_elements( cPagefoot,';' ))

   Begin Sequence    // trap for printer time-out errors here

   // Establish Printer file
   if lForceToFile
      set printer to ( cPtrname )
   endif

   set device to printer

   // Determine nHowWide
   nHowWide := GetWidth()    // Return the appropriate Width
                             // based on Static Variable nWidth

   // check margin problems -- switch to 132 column mode here
   if nWidth + nMargin > 78 .and. ! lPageSkip .and. ;
      aUserInfo[ USER_PRT_PORT ] <>'5' .and. ! lForceToFile
      @ 0,0 say one32()
   end

   if ! lPageSkip
      Pagehead( cPageHead, nWidth, o, nOldMargin )
   else
      nCurLine = 0         // FW-STATIC Variable
   endif

   set( _SET_MARGIN, nMargin )

   nFormSize = formlength() // !!!!!!!

   // For all TBrowse row movements - use the blocks provided
   // with the object ( gotopblock, gobottomblock, skipblock )
   // that way we will solve two problems, 1.  We obey the objects
   // movement definition and scoping rules, 2. We will be able to
   // print any type of data ( not just dbfs )

   eval( o:gotopblock, o )     // go to the top of the data set

   // If Totaling - Set all needed variables for sub total breaks
   If lTotalFlag
      SetBreakValues( o, aBreakColumns )
   Endif

   while nSkipped <> 0      // continue until the high range is reached

      // If Totaling - Check for a control break
      If lTotalFlag
         LookforBreak( o, aBreakColumns )
      Endif

      // draw box characters
      if lBox_Chars
         nThisMargin := set( _SET_MARGIN, 0 )
         @ nCurLine,0 say substr( cBoxChars, 5,1 )
         set( _SET_MARGIN, nThisMargin )
      end

      // Loop Through the Columns of the current row
      for i := 1 to o:colCount
         // check for Column object Decimal positions
         xDecWidth = datadict( o:getcolumn( i ), DECIMALS)

         If aColTypes[i] <> 'N'  // Non-Numeric
            // Non-Numeric data can not be totaled at this time
            @ nCurLine, aOffset[ i ] say ;
                        eval( o:getcolumn( i ):block )
         Else
            w = o:colWidth( i )
            d = eval( o:getcolumn( i ):block )
            temp = str( d, w, if( xDecWidth <> nil, ;
                                  xDecWidth,nDefWidth))
            @ nCurLine, aOffset[ i ] say temp

            // Accumulate Totals if this column is to be totaled
            // NOTE:  Again we use the ACLONE() function because
            //        Clipper handles all array passing as
            //        referential
            If ! Empty( DataDict( o:GetColumn( i ), "COL_TOTAL" ) )
               aTotals := aClone( DataDict( o:GetColumn( i ), ;
                                            "COLUMN_TOTALS" ) )
               for j := 1 to ( len( aBreakColumns ) + 1 )
                  aTotals[ j ] += d       // Increment total array by Cell value
               next

               // Update the column's cargo slot with the new
               // total information
               DataDict( o:GetColumn( i ), "COLUMN_TOTALS", ;
                           aClone( aTotals ) )

            Endif
         Endif
      Next

      // draw box characters
      if lBox_Chars
         nThisMargin := set( _SET_MARGIN, 0 )

         n_colpos := GetWidth()    // Return the appropriate Width
                                   // based on Static Variable nWidth

         @ nCurLine, n_colpos say substr( cBoxChars, 5,1 )
         set( _SET_MARGIN, nThisMargin )
      endif

      nCurLine ++          // FW-STATIC

      // Move to the next row - use the skip block to
      // obey any programmed movement of the TBrowse object

      // The skipblock will return the number of rows moved
      // Skipblock will also return 0 when trying to go beyond
      // the top or bottom of the TBrowse object - Since we know
      // we will not go beyond the top - when nSkipped == 0, we must
      // have tried to go beyond the end of the data - That is why
      // the main processing loop is WHILE nSkipped <> 0.

      nSkipped = eval( o:skipblock, 1 )

      if ! lPageSkip
         if ( nCurLine + nAFooterHeight + 2 ) == nFormSize .and. ;
                     nSkipped <> 0

            Pagefoot( cPageFoot, nWidth, o, .f. )
            nCurLine ++                // FW-Static
            @ nCurLine, 0 say substr( cBoxChars, 3,1 ) + ;
                        repl(substr(cBoxChars,6,1), nHowWide-1) +;
                              substr(cBoxChars,4,1)
            Pagehead( cPageHead, nWidth, o, nOldMargin )

         endif
      endif

      if inkey() == K_ESC .and. nSkipped > 0
         while inkey() <> 0
         end
         nExitState = Alert('Terminate Report;;Are you sure', {'No','Yes'} )
         if nExitState = 2
            exit
         endif
      endif

   end

   if nExitState = 2
      nCurLine += 4
      center( nCurLine, 'Output terminiated by operator', nHowWide)
      nCurLine += 2
   endif
   if nCurLine > nFormSize - ( nFooterHeight + nAFooterHeight + 2 )
      nCurLine = nFormSize - ( nFooterHeight + nAFooterHeight + 2 )
   endif

   // If Totaling - Print all subtotal lines
   If lTotalFlag
      PrtSubTotals( o, 1, aBreakColumns )
   Endif

   if ! lPageSkip
      Pagefoot( cPageFoot, nWidth, o, .t. )
      nThisMargin := set( _SET_MARGIN, 0 )

      while  (nCurLine + 1)  < nFormSize .and. lBox_Chars
         @ nCurLine,0 say substr( cBoxChars, 5,1 )
         @ nCurLine, nHowWide say substr( cBoxChars, 5,1 )
         nCurLine ++
      end
      @ nCurLine,0 say substr( cBoxChars, 3,1 ) + ;
         repl(substr(cBoxChars,6,1), nHowWide-1) + ;
               substr(cBoxChars,4,1)
      nCurLine ++                // FW-Static
      set( _SET_MARGIN, nThisMargin )
   endif

   if ! lForceToFile
      @ 0,0 say Eighty()  // Switch to 80 column mode here!
   end

   // general cleanup

   set printer to  // close the print file

   End Sequence      // end printer time-out here

   rowHeading( -1 )  // kill the static arrays
   rowFooting( -1 )  // kill the static arrays
   set( _SET_MARGIN, nOldMargin )  // reset the margin
   set( _SET_DEVICE, xOldDevice )  // reset the device
   setcursor( xOldCursor )    // reset the cursor
   set( _SET_ESCAPE, xOldEscape )

   // Restore the TBrowse screen
   RestScreen( 0, 0, maxrow(), maxcol(), cScreen )

   while inkey() <> 0; end    // eat any stray key strokes

   o:gotop()
   o:invalidate()       // cause the object to go to the top
                        // of the screen
Return Nil


//************************************************************************
Static Function PageCount( value )
   Static nCount := 0

   if value <> nil  .and. value == 0
      nCount := 0
   endif

   if value == nil
      nCount ++
   endif

Return nCount

//************************************************************************
Static Function RowHeading( o, maxheight )
   Static aRay
   Local i, j, k, cText, aTemp, nNumColSep

   if o <> nil .and. maxheight == 0
      aRay = {}
   endif

   if valtype( o ) == 'N' .and. o = -1
      // kill the static array
      aRay = nil
   endif

   // fill up the array with blanks
   if maxheight <> nil  .and. maxheight > 0
      aTemp := array( maxheight, o:colCount )
      for i = 1 to maxheight
         for j = 1 to o:colCount
            aTemp[ i,j ] = space( o:colwidth( j ) )
         next
      next
   endif

   if o <> nil .and. valtype( o ) == 'O'.and. maxheight > 0

      for j = 1 to o:colCount
         i = maxheight

         // rParse() is defined below
         cText = rParse( o:getColumn( j ):heading, ';')

         while len( cText ) > 0
            cText = left( cText, o:ColWidth( j ) )
            aTemp[ i,j ] = padr( cText, o:ColWidth( j ) )
            cText = rParse()
            i --
         end
      next
   endif

   if maxheight <> nil .and. maxheight > 0
      nNumColSep = datadict( o, COLSEPS )
      default nNumColSep to 1
      aRay := array( maxheight )

      for i = 1 to maxheight
         aRay[ i ] = if(lBox_Chars, '  ', '')

         for j = 1 to o:colCount
            aRay[ i ] += aTemp[i,j]  + space( nNumColSep )
         next
         aRay[ i ] = trim(aRay[ i ])
      next
   endif

Return aRay

//************************************************************************
Static Function RowFooting( o, maxheight )
   Static aRay
   Local i, j, k, cText, aTemp, nNumColSep

   if o <> nil .and. maxheight == 0
      aRay = {}
   endif

   if valtype( o ) == 'N' .and. o = -1
      // kill the static array
      aRay = nil
   endif

   // fill up the array with blanks
   if maxheight <> nil .and. maxheight > 0
      aTemp := array( maxheight, o:colCount )

      for i = 1 to maxheight
         for j = 1 to o:colCount
            aTemp[ i,j ] = space( o:colwidth( j ) )
         next
      next
   endif

   if o <> nil .and. valtype( o ) == 'O' .and. maxheight > 0
      for j = 1 to o:colCount
         i = 1

         // Parse() is defined below
         cText = parse( o:getColumn( j ):footing, ';')

         while len( cText ) > 0
            cText = left( cText, o:ColWidth( j ) )
            aTemp[ i,j ] = padr( cText, o:ColWidth( j ) )
            cText = parse()
            i ++
         end
      next
   endif

   if !( maxheight == nil ) .and. maxheight > 0
      nNumColSep = datadict( o, COLSEPS )
      default nNumColSep to 1
      aRay := array( maxheight )

      for i = 1 to maxheight
         aRay[ i ] = if(lBox_Chars, '  ', '')

         for j = 1 to o:colCount
            aRay[ i ] += aTemp[i,j] + space( nNumColSep )
         next
         aRay[ i ] = trim(aRay[ i ])
      next
   endif

Return aRay

//************************************************************************
Static Function PageHead( cHeading, nWidth, o, nOldMargin )
   local cText, i, aRowHead := rowHeadings(), ;
         nMax, nThisMargin

   // Alters the FW-Static variable nCurLine   !!!!!!!!!!!!!!!1

   default cHeading to '', ;
           nWidth to 80, ;
           nOldMargin to 0

   if lBox_Chars
      @ 0,0 say left( cBoxChars, 1) + ;
                repl(substr(cBoxChars,6,1),nHowWide-1) + ;
                     substr(cBoxChars,2,1)
   endif

   @ 1,0 say substr( cBoxChars, 5,1 )
   @ 1,1 say date() pict '@D'
   @ 1,nHowWide say substr( cBoxChars, 5,1 )

   @ 2,0 say substr( cBoxChars, 5,1 )
   @ 2,1 say 'Page ' + ltrim( str( pagecount(),4,0 ) )
   @ 2,nHowWide say substr( cBoxChars, 5,1 )

   nCurLine = 3

   set( _SET_MARGIN, nOldMargin )

   cText = Parse( cHeading, ';' )

   while len( cText ) > 0
      @ nCurLine, 0 say substr( cBoxChars, 5,1 )
      center( nCurLine, cText, nHowWide )
      @ nCurLine, nHowWide say substr( cBoxChars, 5,1 )
      cText = Parse()
      nCurLine ++
   end

   @ nCurLine, 0 say substr( cBoxChars, 5,1 )
   @ nCurLine, nHowWide say substr( cBoxChars, 5,1 )

   nCurLine ++

   set( _SET_MARGIN, nMargin )

   for i = 1 to len( aRowHead )
      nThisMargin := set( _SET_MARGIN, 0 )
      @ nCurLine,0 say substr( cBoxChars, 5,1 )
      set( _SET_MARGIN, nThisMargin )

      @ nCurLine, 0 say substr( aRowHead[ i ], 1, ( nHowWide - 1 ) )

      nThisMargin := set( _SET_MARGIN, 0 )
      @ nCurLine,nHowWide say substr( cBoxChars, 5,1 )
      set( _SET_MARGIN, nThisMargin )

      nCurLine ++
   next

   nThisMargin := set( _SET_MARGIN, 0 )
   @ nCurLine,0 say substr( cBoxChars, 5,1 )
   set( _SET_MARGIN, nThisMargin )

   for i = 1 to o:colCount
      // double box line if char set permits
      @ nCurLine, aOffset[i] say;
         repl( chr( if( aUserInfo[ USER_CHAR_SET ] = '1', ;
                        asc('-'), 205 )),;
         o:colWidth(i) )
   next

   nThisMargin := set( _SET_MARGIN, 0 )
   @ nCurLine, nHowWide say substr( cBoxChars, 5,1 )
   set( _SET_MARGIN, nThisMargin )

   nCurLine ++

Return Nil

//************************************************************************
Static Function PageFoot( cFooting, nWidth, o, lPrintFooters)
   local cText, i, aRowFoot := rowFootings(), nThisMargin

   default cFooting to '', ;
           nWidth to 80, ;
           lPrintFooters to .f.

   if lPrintFooters
      if ! empty( aRowFoot )
         nThisMargin := set( _SET_MARGIN, 0 )
         @ nCurLine,0 say substr( cBoxChars, 5,1 )
         set( _SET_MARGIN, nThisMargin )

         for i = 1 to o:colCount
            // double box line
            @ nCurLine, aOffset[i] say;
                     repl( chr( if( aUserInfo[ USER_CHAR_SET ] = '1', ;
                           asc('-'), 205 ) ), o:colWidth(i) )
         next

         nThisMargin := set( _SET_MARGIN, 0 )
         @ nCurLine,nHowWide say substr( cBoxChars, 5,1 )
         set( _SET_MARGIN, nThisMargin )

         nCurLine ++          // FW-Static
      endif

      for i = 1 to len( aRowFoot )

         nThisMargin := set( _SET_MARGIN, 0 )
         @ nCurLine,0 say substr( cBoxChars, 5,1 )
         set( _SET_MARGIN, nThisMargin )

         @ nCurLine, 0 say substr( aRowFoot[ i ], 1, ( nHowWide - 1 ) )

         nThisMargin := set( _SET_MARGIN, 0 )
         @ nCurLine,nHowWide say substr( cBoxChars, 5,1 )
         set( _SET_MARGIN, nThisMargin )

         nCurLine ++          // FW-Static
      next
   endif

   set( _SET_MARGIN, 0 )
   @ nCurLine,0 say substr( cBoxChars, 5,1 )
   @ nCurLine,nHowWide say substr( cBoxChars, 5,1 )

   cText = Parse( cFooting, ';' )

   if ! empty( cText )
      nCurLine ++             // FW-Static

      while len( cText ) > 0
         @ nCurLine,0 say substr( cBoxChars, 5,1 )
         center( nCurLine, cText, nHowWide )
         @ nCurLine,nHowWide say substr( cBoxChars, 5,1 )

         cText = Parse()
         nCurLine ++          // FW-Static
      end
      nCurLine --             // Ditto
   endif

Return Nil

/////////////////////////////////////////////////////////////////////////
Function PtrName
   Local cName, nCount := 0, nHandle, lOk := .f., cCount, retval := ''

   while ! lOk .and. nCount < 100
      cName = trim( aUserInfo[ USER_NAME ] )
      cCount = right( '0000000'+ltrim(str( ++ nCount, 3, 0 )),8-len(cName) )
      nHandle = fopen( cName + cCount + '.RPT')

      if nHandle = -1      // the file is not there
         retval = cName + cCount + '.RPT'
         exit
      else
         fclose( nHandle )
      endif
   end

Return retval

/////////////////////////////////////////////////////////////////////////
Static Function GetWidth()
   Local nRealWidth

   // Based on the value of the FW-Static variable nWidth
   // Return the actual printing width of the paper
   Do Case
      Case nWidth < 79
         nRealWidth := 79
      Case nWidth > 78 .and. nWidth < 133
         nRealWidth := 126
      Case nWidth > 133
         nRealWidth := 178
   EndCase

Return nRealWidth

///////////////////////////////////////////////////////////////////////
Static Function One32
   // Set static variable nWidth to paper width and
   // returns to the appropriate printer setup string

   local retval := chr( 15 )  // assume dot matrix

   //If options->formwidth > 80
   //   retval = chr( 18 )   // no action
   //Endif

   nWidth = 132
   If aUserInfo[ USER_PRT_TYPE ] == 'L'
      lInLandScape = .t.
      If nHowWide < 127
         retval = ESC + "&l1O" + ESC + "(s0p12H"   // 12 Pitch - Normal Land
         nWidth = 126
      Else
         retval = ESC + "&l1O" + ESC + ;
                  "(s0p16.6H"                      // 16.6 Pitch - Condensed Land
         nWidth = 178
      Endif
   Endif

Return retval

/////////////////////////////////////////////////////////////////////////////
Static Function Eighty
   local cRetval := chr( 18 )  // assume dot matrix

   if aUserInfo[ USER_PRT_TYPE ] == 'L'
      cRetval = ESC + "&l0O" + ESC + "(s0p10H"      // Normal Port
   endif

Return cRetval

///////////////////////////////////////////////////////////////////////////
Static Function FormLength
   local nRetval := 58

   If lInLandScape
      nRetval =  43
   endif

Return nRetval

////////////////////////////////////////////////////////////////////////
Static Function SetBreakValues( o, aBreakColumns )
   Local aBreaks := {}, i, xBlock

   // If there are break levels
   If ValType( aBreakColumns ) == 'A'

      For i := 1 To len( aBreakColumns )
         // aBreakColumns is an array of columns numbers
         // in their proper control break order - if the programmer
         // has defined a codeblock as a break condition - use it
         // otherwise use the contents of the cell
         xBlock := DataDict( o:GetColumn( aBreakColumns[ i ] ), "COL_BREAK" )

         // If the cargo slot COL_BREAK has a code block - use the code
         // block to determine break values - otherwise break upon
         // a change in the cell
         If ValType( xBlock ) == 'B'
            aadd( aBreaks, eval( xBlock ) )
         Else
            aadd( aBreaks, eval( o:GetColumn( aBreakColumns[ i ] ):block ) )
         Endif
      Next

      // Copy this array into the FW-Static variable aSBreaks
      aSBreaks := aClone( aBreaks )
   Endif

Return Nil

////////////////////////////////////////////////////////////////////////
Static Function LookforBreak( o, aBreakColumns )
   Local i, xBlock, xCompValue

   // If there are break levels
   If ValType( aBreakColumns ) == 'A'

      For i := 1 To len( aBreakColumns )

         // Compare static array to current cell values or block values

         // As stated above - if COL_BREAK has a code block - use the
         // Code block to determine a break - otherwise use the cells contents
         xBlock := DataDict( o:GetColumn( aBreakColumns[ i ] ), "COL_BREAK" )

         If ValType( xBlock ) == 'B'
            xCompValue := eval( xBlock )
         Else
            xCompValue := eval( o:GetColumn( aBreakColumns[ i ] ):block )
         Endif

         If aSBreaks[ i ] <> xCompValue
            // We have a break - so call the routine
            // that prints the subtotals
            // Sub Totals are stored in an array
            // that is one larger than len( aSBreaks )
            // This is due to the grand total array element.
            // Therefore we pass i + 1 to PrtSubTotals so
            // we can point to the correct array element

            PrtSubTotals( o, i + 1, aBreakColumns )
         Endif

      Next

   Endif

Return Nil

//////////////////////////////////////////////////////////////////////
Static Function PrtSubTotals( o, nHighLevel, aBreakColumns )
   Local aStrings := datadict( o, "TB_STRINGS" ), j, i, ;
         w, d, temp, aTotals, xDecWidth, n_ColPos, ;
         nDefWidth := datadict( o, DECIMALS ), nThisMargin
   Local cPageHead, cPageFoot, cSubString, nPoundPos

   Default nDefWidth to 0

   // Extract the Page Header and Page Footer
   cPageHead = datadict( o, 'PAGEHEADER' )
   cPageFoot = datadict( o, 'PAGEFOOTER' )

   // Start at the lowest break level and work up to
   // the value of i ( from LookforBreak() ) known as nHighLevel
   // in this routine.  For each level print the subtotal line.

   For j := Len( aBreakColumns ) + 1 To nHighLevel Step -1

      // blank line before subtotal line
      if lBox_Chars
         nThisMargin := set( _SET_MARGIN, 0 )
         @ nCurLine,0 say substr( cBoxChars, 5,1 )
         set( _SET_MARGIN, nThisMargin )
      endif

      if lBox_Chars
         nThisMargin := set( _SET_MARGIN, 0 )

         n_colpos := GetWidth()    // Return the appropriate Width
                                   // based on Static Variable nWidth

         @ nCurLine, n_colpos say substr( cBoxChars, 5,1 )
         set( _SET_MARGIN, nThisMargin )
      endif

      nCurLine ++          // FW-Static

      // draw box characters
      if lBox_Chars
         nThisMargin := set( _SET_MARGIN, 0 )
         @ nCurLine,0 say substr( cBoxChars, 5,1 )
         set( _SET_MARGIN, nThisMargin )
      endif

      // Print the subtotal's literal string - this comes from an
      // array in the TBrowse's cargo slot titled: TB_STRINGS
      If len( aStrings ) > 0
         cSubString := aStrings[ j ]

         // Embedded ## in the TB_STRING means to place the value
         // of the break column in the subtotal literal string
         nPoundPos := at( '##', cSubString )
         If nPoundPos > 0
            // Replace the ## with the current break value
            cSubString := strTran( cSubString, "##", aSBreaks[ j - 1 ] )
         endif
         @ nCurLine, 0 say cSubString
      Endif

      for i := 1 to o:colCount
         xDecWidth = datadict( o:getcolumn( i ), DECIMALS)

         // if the column is numeric AND is defined to be totaled...
         If valtype( eval( o:getcolumn( i ):block ) ) == 'N' .and. ;
                  ! Empty( DataDict( o:GetColumn( i ), "COL_TOTAL" ) )

            // Extract the total array from the columns cargo slot
            aTotals := aClone( DataDict( o:GetColumn( i ), "COLUMN_TOTALS" ) )

            w = o:colWidth( i )
            d = aTotals[ j ]        // Get the subtotal element of this loop
            temp = str( d, w, if( xDecWidth <> nil, ;
                                  xDecWidth,nDefWidth))
            @ nCurLine, aOffset[ i ] say temp

            // Reset the element to zero
            aTotals[ j ] := 0

            // Replace the array into the column's cargo slot
            // remember to use aclone()!
            DataDict( o:GetColumn( i ), "COLUMN_TOTALS", aClone( aTotals ) )
         Endif
      Next

      // draw box characters
      if lBox_Chars
         nThisMargin := set( _SET_MARGIN, 0 )

         n_colpos := GetWidth()    // Return the appropriate Width
                                   // based on Static Variable nWidth

         @ nCurLine, n_colpos say substr( cBoxChars, 5,1 )
         set( _SET_MARGIN, nThisMargin )
      endif

      nCurLine ++          // FW-Static

      if ! lPageSkip
         if ( nCurLine + nAFooterHeight + 2 ) >= nFormSize

            Pagefoot( cPageFoot, nWidth, o, .f. )
            nCurLine += 1      // ... cwk
            @ nCurLine, 0 say substr( cBoxChars, 3,1 ) + ;
                        repl(substr(cBoxChars,6,1), nHowWide-1) +;
                              substr(cBoxChars,4,1)
            Pagehead( cPageHead, nWidth, o )

         endif
      endif

   Next

   // blank line after all subtotal lines
   if lBox_Chars
      nThisMargin := set( _SET_MARGIN, 0 )
      @ nCurLine,0 say substr( cBoxChars, 5,1 )
      set( _SET_MARGIN, nThisMargin )
   endif

   if lBox_Chars
      nThisMargin := set( _SET_MARGIN, 0 )
      n_colpos := GetWidth()    // Return the appropriate Width
                                // based on Static Variable nWidth

      @ nCurLine, n_colpos say substr( cBoxChars, 5,1 )
      set( _SET_MARGIN, nThisMargin )
   endif

   nCurLine ++                      // FW-Static

   SetBreakValues( o, aBreakColumns )     // Reset the break column values

Return Nil

///////////////////////////////////////////////////////////////////////////
// Store Employee print information for TBPrint()
// This function was designed for a specific application
// For your use you can hardcode the options here or store
// them in a database as the commented code shows

Function SetUserInfo
// The employee printer type could be stored in a database
// containing employee specific information.

   aUserInfo := array( USER_INFO_COUNT )
   aUserInfo[ USER_PRT_TYPE ] = 'D'         // Dot Matrix
   aUserInfo[ USER_PRT_PORT ] = '1'         // LPT1
   aUserInfo[ USER_CHAR_SET ] = '2'         // Full character set
   aUserInfo[ USER_NAME ]     = 'EXAMPL'    // Default file Name

Return aUserInfo

// ************************************************************************
Function DataDict( obj, cRowName, datam )
   local nIndex, retval := nil

   // If the objects cargo slot is not in use - initialize it as an array
   if obj:cargo == nil
      obj:cargo = {} // force an empty array into the cargo slot
   endif

   // Locate the cRowName in the Cargo array
   nIndex = ascan( obj:cargo, { | elem | upper(elem[1]) == upper(cRowName) } )

   // If the cRowName was not found and datam was passed
   // dynamically grow cargo and insert the cRowName and the datam
   if nIndex == 0 .and. ! ( datam == nil )
      aadd( obj:cargo, { cRowName, datam } )
      nIndex = len( obj:cargo )
   endif

   // If the cRowName was found and datam was passed
   // Store the datam into cargo
   if nIndex <> 0
      if !( datam == nil)
         obj:cargo[ nIndex, 2 ] = datam
      endif

      // Return the value of the cargo slot requested by cRowName
      retval = obj:cargo[ nIndex, 2 ]
   endif

Return retval

//////////////////////////////////////////////////////////////////////
/*
   Parse a string from the right to the left

    Usage:
          cText := rParse( 'This;is;a;test', ';')
          while ! empty( cText )
                ? cText
                cText := rParse()
          enddo

      OutPut:  test
               a
               is
               This
                                    */

Function rparse( input_text, input_delimiter )
   Static text_to_parse, delimiter := ' '
   Local x, retval

   // If text was passed - store it in the static variable
   if input_text <> NIL
      text_to_parse = input_text
   endif

   // If a delimiter was passed - store it to the static variable
   if input_delimiter <> nil
      delimiter = input_delimiter
   endif

   x = rat(delimiter, text_to_parse)

   if x = 0
      // no more spaces - end of text
      retval = text_to_parse
      text_to_parse = ''
   else
      retval = substr(text_to_parse, x + 1)

      if len( retval ) == 0
         retval = ' '
      endif

      text_to_parse = left( text_to_parse, x-1 )
   endif

Return retval


//////////////////////////////////////////////////////////////////////
/*
   Parse a string from the left to the right

    Usage:
          cText := Parse( 'This;is;a;test', ';')
          while ! empty( cText )
                ? cText
                cText := Parse()
          enddo

      OutPut:  This
               is
               a
               test
                                    */
Function parse ( input_text, input_delimiter )
   Static text_to_parse, delimiter := ' '
   Local x, retval

   if input_text <> NIL
      text_to_parse = input_text
   endif

   if input_delimiter <> nil
      delimiter = input_delimiter
   endif

   x = at(delimiter, text_to_parse)

   if x = 0
      // no more spaces - end of text
      retval = text_to_parse
      text_to_parse = ''
   else
      retval = left(text_to_parse, x - 1)

      if len( retval ) == 0
         retval = ' '
      endif
      text_to_parse = substr( text_to_parse, x+1 )
   endif

Return retval

// **************************************************
Function center( nRow, cTextToCenter, nWidth , cColor)

/*  The ubiquitous center function
    Usage:	center( <row>, <message>, [<width_other_than_80_wide>] )
*/

   Local nCenter

   Default nWidth to 80,;
           cColor to setcolor()

   nCenter = int(nWidth * .5) - int(len(cTextToCenter) * .51)
   @ nRow, nCenter   say cTextToCenter color cColor

Return nCenter

/////////////////////////////////////////
Function how_many_text_elements( heading, delimiter )
/* 	Return the the length of the longest heading field for a tbrowse

	Example:

	how_many_text_elements('I.D.;Number', ';')	=> 	2
*/

   Local cnt :=0, x

   Default delimiter to ';'

   // seed the parse function
   x = parse( heading, delimiter)

   do while len(x) > 0
      cnt ++
      // get the next heading line
      x = parse()
   enddo

Return cnt

// {EOF}
